/*===========================================================================
 *
 * File:	Dfzcfg.CPP
 * Author:	Dave Humphrey (uesp@m0use.net)
 * Created On:	January 18, 2003
 *
 * Description
 *
 *=========================================================================*/

	/* Include Files */
#include "dfzcfg.h"


/*===========================================================================
 *
 * Begin Local Definitions
 *
 *=========================================================================*/
  DEFINE_FILE("DFZCfg.cpp");
/*===========================================================================
 *		End of Local Definitions
 *=========================================================================*/


/*===========================================================================
 *
 * Class CDFZConfig Constructor
 *
 *=========================================================================*/
CDFZConfig::CDFZConfig() {
 m_NumRecords = 0;
 }
/*===========================================================================
 *		End of Class CDFZConfig Constructor
 *=========================================================================*/


/*===========================================================================
 *
 * Class CDFZConfig Method - void Destroy (void);
 *
 *=========================================================================*/
void CDFZConfig::Destroy (void) {
  int Index;

	/* Unallocate all records */
  for (Index = 0; Index < m_NumRecords; Index++) {
    DestroyArrayPointer(m_Records[Index].pVariable);
    DestroyArrayPointer(m_Records[Index].pValue);
   }

  m_NumRecords = 0;
  m_File.Destroy();
 }
/*===========================================================================
 *		End of Class Method CDFZConfig::Destroy()
 *=========================================================================*/


/*===========================================================================
 *
 * Class CDFZConfig Method - bool AddConfig (pVariable, pValue);
 *
 * Adds a new config record to the end of the array. Returns false on
 * any error.
 *
 *=========================================================================*/
bool CDFZConfig::AddConfig (const char* pVariable, const char* pValue) {

	/* Check for array overloads */
  if (m_NumRecords >= DFZCFG_MAX_LINES) {
    ErrorHandler.AddError(ERR_MAXINDEX, "Exceeded the number of Z.CFG records (%d)!", DFZCFG_MAX_LINES);
    return (false);
   }

  m_Records[m_NumRecords].pVariable = CreateString(pVariable);
  m_Records[m_NumRecords].pValue    = CreateString(pValue);
  m_NumRecords++;

  return (true);
 }
/*===========================================================================
 *		End of Class Method CDFZConfig::AddConfig()
 *=========================================================================*/


/*===========================================================================
 *
 * Class CDFZConfig Method - bool CreateDefault (void);
 *
 * Fills the config file with a default set of parameters.
 *
 *=========================================================================*/
bool CDFZConfig::CreateDefault (void) {
  bool Result;
  Destroy();

  Result  = AddConfig("type",		"dfall_huge");
  Result &= AddConfig("path",		"dagger\\arena2\\");
  Result &= AddConfig("pathCD",		"dagger\\arena2\\");
  Result &= AddConfig("fadecolor",	"0");
  Result &= AddConfig("mapfile",	"d");
  Result &= AddConfig("rendergame",	"1");
  Result &= AddConfig("user",		"1");
  Result &= AddConfig("startMap",	"179");
  Result &= AddConfig("region",		"17");
  Result &= AddConfig("helmet",		"0");
  Result &= AddConfig("maxSpeed",	"200");
  Result &= AddConfig("controls",	"betaplyr.dat");
  Result &= AddConfig("maps",		"mapsave.sav");
  Result &= AddConfig("cheatmode",	"1");

  return (Result);
 }
/*===========================================================================
 *		End of Class Method CDFZConfig::CreateDefault()
 *=========================================================================*/


/*===========================================================================
 *
 * Class CDFZConfig Method - int FindRecord (pVariable);
 *
 * Returns the first index that matches the given variable.  Returns -1 
 * on any error.
 *
 *=========================================================================*/
int CDFZConfig::FindRecord (const char* pVariable) { 
  int Index;
  int Result;

	/* Ignore invalid input */
  if (pVariable == NULL) return (-1);

	/* Search entire array */
  for (Index = 0; Index < m_NumRecords; Index++) {
    if (m_Records[Index].pVariable == NULL) continue;

    Result = stricmp(m_Records[Index].pVariable, pVariable);
    if (Result == 0) return (Index);
   }

	/* No match found */
  return (-1);    
 }
/*===========================================================================
 *		End of Class Method CDFZConfig::FindRecord()
 *=========================================================================*/


/*===========================================================================
 *
 * Class CDFZConfig Method - bool Load (pFilename, MakeDefault);
 *
 * Attempt to load the given Z.CFG for input.  If MakeDefault is true and
 * the Z.CFG file does not exist, default values will be created instead.
 * Returns false on any error.
 *
 *=========================================================================*/
bool CDFZConfig::Load (const char* pFilename, const bool MakeDefault) {
  bool  Result;
  int   ReadResult = READLINE_OK;
  char  Buffer[DFZCFG_LINE_LENGTH+1];
  char* pVariable;
  char* pValue;

	/* Delete the current contents */
  Destroy();

	/* Attempt to open the file for input */
  Result = m_File.Open(pFilename, "rt");

  if (!Result) {
    if (MakeDefault) return CreateDefault();
    return (false);
   }

	/* Input and parse the entire file */
  while(!m_File.IsEOF()) {
    ReadResult = m_File.ReadLine(Buffer, DFZCFG_LINE_LENGTH);
    if (ReadResult == READLINE_ERROR) break;

    pVariable = trim(Buffer);
    pValue = strchr(pVariable, ' ');
    if (pValue == NULL) pValue = strchr(pVariable, '\t');
    
    if (pValue != NULL) {
      *pValue = NULL_CHAR;
      pValue = ltrim(pValue + 1);
      rtrim(pVariable);
     }

    AddConfig(pVariable, pValue);
   }

  m_File.Close();
  return (ReadResult != READLINE_ERROR);
 }
/*===========================================================================
 *		End of Class Method CDFZConfig::Load()
 *=========================================================================*/


/*===========================================================================
 *
 * Class CDFZConfig Method - bool Save (pFilename);
 *
 *=========================================================================*/
bool CDFZConfig::Save (const char* pFilename) {
  bool  Result;
  int   Index;

	/* Attempt to open the file for output */
  Result = m_File.Open(pFilename, "wt");
  if (!Result) return(false);

	/* Input and parse the entire file */
  for (Index = 0; Index < m_NumRecords; Index++) {

    if (m_Records[Index].pValue == NULL) 
      Result = m_File.Printf ("%s\n", m_Records[Index].pVariable);
    else
      Result = m_File.Printf ("%s %s\n", m_Records[Index].pVariable, m_Records[Index].pValue);
  
    if (!Result) {
      m_File.Close();
      return (false);
     }  
   }

  m_File.Close();
  return (true);
 }
/*===========================================================================
 *		End of Class Method CDFZConfig::Save()
 *=========================================================================*/


/*===========================================================================
 *
 * Class CDFZConfig Method - bool SetValue (pVariable, pValue);
 *
 * Changes the value of the given variable, or adds it if it does not 
 * exist.
 *
 *=========================================================================*/
bool CDFZConfig::SetValue (const char* pVariable, const char* pValue) {
  int Result;

  Result = FindRecord(pVariable);
  if (Result < 0)  return (AddConfig(pVariable, pValue));
   
  DestroyArrayPointer(m_Records[Result].pValue);
  m_Records[Result].pValue = CreateString(pValue);
  return (true);
 }
/*===========================================================================
 *		End of Class Method CDFZConfig::SetValue()
 *=========================================================================*/
